/* -*- mode: c; c-set-style: "stroustrup"; tab-width: 4; buffer-file-coding-system: unix; -*-
 *
 * main.c
 *
 * Copyright (C) 2005 Koji Nakamaru
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include <fcntl.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "error.h"

#define k_app_name			"psfindpages"
#define k_version			"1.00"
#if defined(_WIN32) && !defined(__CYGWIN__)
#define k_gs_default		"gswin32c"
#define k_dev_null			"NUL"
#else
#define k_gs_default		"gs"
#define k_dev_null			"/dev/null"
#endif
#define k_gs_args			" -dBATCH -dNOPAUSE -r72 -dTextAlphaBits=4 -dGraphicsAlphaBits=4 -sDEVICE=ppmraw -sstdout=" k_dev_null " -sOutputFile=- -"
#define k_type_default		"cg"
#define k_col_bits			(sizeof(*_col) * 8)
#define k_col_inc			(100)

static void parseOptions(int argc, char *argv[]);
static void usage();
static void checkPages();
static int readHeader(int *xres, int *yres);
static void initCol();
static void addCol(int i);
static void output();
static void outputInterval(int *is_first, int s, int e);

static struct option _options[] = {
	{ "gs", 1, NULL, 'g' },
	{ "type", 1, NULL, 't' },
	{ "help", 0, NULL, 'h' },
	{ "version", 0, NULL, 'v' },
	{ NULL, 0, NULL, 0 },
};
static char *_gs = k_gs_default;
static char *_type = k_type_default;
static FILE *_pp;
static int *_col;
static int _col_size;
static int _col_n;

int main(
	int argc,
	char *argv[])
{
	char buf[BUFSIZ];
#if defined(_WIN32) && !defined(__CYGWIN__)
	_fmode = _O_BINARY;
	_setmode(0, _O_BINARY);
	_setmode(1, _O_BINARY);
	_setmode(2, _O_BINARY);
#endif
	setProgname(argv[0]);
	parseOptions(argc, argv);
	if (strlen(_gs) + strlen(k_gs_args) + 1 > BUFSIZ) {
		error("cannot accept too long gs name - '%s'", _gs);
	}
	strcpy(buf, _gs);
	strcat(buf, k_gs_args);
	if ((_pp = popen(buf, "r")) == NULL) {
		error("failed to invoke '%s'", _gs);
	}
	checkPages();
	pclose(_pp);
	if (strchr(_type, 'c')) {
		output(1);
	}
	if (strchr(_type, 'g')) {
		output(0);
	}
	return 0;
}

static void parseOptions(
	int argc,
	char *argv[])
{
	int c;
	opterr = 0;
	optind = 1;
	while ((c = getopt_long(argc, argv, "g:t:hv", _options, NULL)) != -1) {
		switch (c) {
		case 'g':
			_gs = optarg;
			break;
		case 't':
			if (strcmp(optarg, "c") == 0
				|| strcmp(optarg, "g") == 0
				|| strcmp(optarg, "cg") == 0) {
				_type = optarg;
			} else {
				usage();
			}
			break;
		case 'v':
			fprintf(
				stderr,
				"%s %s\n"
				"Copyright (C) 2005 Koji Nakamaru.\n"
				"This is free software; see the source for copying conditions.  There is NO\n"
				"warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.\n",
				k_app_name,
				k_version);
			exit(1);
		case 'h':
		default:
			usage();
		}
	}
}

static void usage()
{
	fprintf(
		stderr,
		"%s %s\n"
		"Usage: %s [OPTION]...\n"
		"shows which pages are gray/color.\n"
		"  -g, --gs=GS               use GS as the command name of ghostscript (default %s).\n"
		"  -t, --type=TYPE           use TYPE (c, g, or cg) to determine which types of\n"
		"                            pages to be shown (default: %s).\n"
		"  -h, --help                display this help and exit.\n"
		"  -v, --version             output version information and exit.\n",
		k_app_name,
		k_version,
		progname(),
		k_gs_default,
		k_type_default);
	exit(1);
}

static void checkPages()
{
	char buf[3];
	int xres, yres, x, y;
	int is_color;
	is_color = 0;
	initCol();
	while (readHeader(&xres, &yres)) {
		is_color = 0;
		for (y = 0; y < yres; y++) {
			for (x = 0; x < xres; x++) {
				if (fread(buf, 3, 1, _pp) != 1) {
					error("failed to read from gs.");
				}
				if (buf[0] != buf[1] || buf[1] != buf[2]) {
					is_color = 1;
				}
			}
		}
		addCol(is_color);
	}
	addCol(! is_color); /* sentinel */
}

static int readHeader(
	int *xres,
	int *yres)
{
	char buf[256];
	int maxval;

	if (fread(buf, 3, 1, _pp) != 1) {
		if (feof(_pp)) {
			return 0;
		} else if (memcmp(buf, "P6\n", 3) != 0) {
			error("found unexpected data from gs.");
		}
	}
	/* the following does not fully conform to the specification of ppm */
	if (fgets(buf, sizeof(buf), _pp) == NULL  /* '#' */
		|| fgets(buf, sizeof(buf), _pp) == NULL
		|| sscanf(buf, "%d %d", xres, yres) != 2
		|| fgets(buf, sizeof(buf), _pp) == NULL
		|| sscanf(buf, "%d", &maxval) != 1) {
		error("found unexpected data from gs.");
	}
	return 1;
}

static void initCol()
{
	_col_size = k_col_inc;
	if ((_col = (int *)malloc(sizeof(*_col) * _col_size)) == NULL) {
		error("failed to allocate memory.");
	}
	_col_n = 0;
}

static void addCol(
	int i)
{
	if (_col_n / k_col_bits >= _col_size) {
		int *col;
		_col_size += k_col_inc;
		if ((col = (int *)realloc(_col, sizeof(*_col) * _col_size)) == NULL) {
			error("failed to allocate memory.");
		}
		_col = col;
		memset(&_col[_col_size - k_col_inc], 0, k_col_inc);
	}
	if (i == 1) {
		_col[_col_n / k_col_bits] |= 1 << (_col_n % k_col_bits);
	} else {
		_col[_col_n / k_col_bits] &= ~(1 << (_col_n % k_col_bits));
	}
	_col_n++;
}

static void output(
	int is_color)
{
	int is_first, mode, s, e, i;
	is_first = 1;
	mode = 0;
	s = 1;
	for (i = 0; i < _col_n; i++) {
		e = i;
		if (mode == 0) {
			if (_col[i / k_col_bits] & (1 << (i % k_col_bits))) {
				if (! is_color) {
					outputInterval(&is_first, s, e);
				}
				s = i + 1;
				mode = 1;
			}
		} else {
			if (! (_col[i / k_col_bits] & (1 << (i % k_col_bits)))) {
				if (is_color) {
					outputInterval(&is_first, s, e);
				}
				s = i + 1;
				mode = 0;
			}
		}
	}
	putchar('\n');
}

static void outputInterval(
	int *is_first,
	int s,
	int e)
{
	if (*is_first) {
		*is_first = 0;
	} else {
		putchar(',');
	}
	if (s < e) {
		printf("%d-%d", s, e);
	} else {
		printf("%d", s);
	}
}
